@@ -17,29 +17,42 @@ import Junc
 import AST
 import Pretty
 import Parser
+import Monads
 
-op0 :: Ident -> [Val] -> Val
-op0 ","  = VList . concatMap vCast
-op0 "!"  = VJunc . Junc JNone emptySet . mkSet
-op0 "&"  = opJuncAll
-op0 "^"  = opJuncOne
-op0 "|"  = opJuncAny
-op0 s    = \x -> VError ("unimplemented listOp: " ++ s) (Val $ VList x)
+op0 :: Ident -> [Val] -> Eval Val
+op0 ","  = return . VList . concatMap vCast
+op0 "!"  = return . VJunc . Junc JNone emptySet . mkSet
+op0 "&"  = return . opJuncAll
+op0 "^"  = return . opJuncOne
+op0 "|"  = return . opJuncAny
+op0 "time"  = \_ -> do
+    clkt <- liftIO getClockTime
+    return $ VInt $ toInteger $ tdSec $ diffClockTimes clkt epochClkT
+    where
+    epochClkT = toClockTime epoch
+    epoch = CalendarTime 1970 January 1 0 0 0 0 Thursday 0 "UTC" 0 False
+op0 s    = \x -> return $ VError ("unimplemented listOp: " ++ s) (Val $ VList x)
 
 op1 :: Ident -> Val -> Eval Val
 op1 "!"    = return . fmapVal not
 op1 "+"    = return . op1Numeric id
-op1 "++"   = \mv -> do
+op1 "post:++" = \mv -> do
     val <- readMVal mv
     liftIO $ writeIORef (vCast mv) $ case val of
         (VStr str)  -> VStr $ strInc str
         _           -> op1Numeric (\x -> x + 1) (vCast val)
-    return val
-op1 "--"   = \mv -> do
+    op1 "+" val
+op1 "++"   = \mv -> do
+    op1 "post:++" mv
+    readMVal mv
+op1 "post:--"   = \mv -> do
     val <- liftIO $ readIORef (vCast mv)
     liftIO $ writeIORef (vCast mv) $
         op1Numeric (\x -> x - 1) (vCast val)
-    return val
+    op1 "+" val
+op1 "--"   = \mv -> do
+    op1 "post:--" mv
+    readMVal mv
 op1 "-"    = return . op1Numeric negate
 op1 "~"    = return . VStr . vCast
 op1 "?"    = return . VBool . vCast
@@ -62,12 +75,6 @@ op1 "last" = \v -> do
 op1 "return" = \v -> return (VError "cannot return outside a subroutine" (Val v))
 
 -- Side-effectful function: how far into Monadic IO shall we go?
-op1 "time"  = \v -> do
-    clkt <- liftIO getClockTime
-    return $ VInt $ toInteger $ tdSec $ diffClockTimes clkt epochClkT
-    where
-    epochClkT = toClockTime epoch
-    epoch = CalendarTime 1970 January 1 0 0 0 0 Thursday 0 "UTC" 0 False
 op1 "rand"  = \v -> do
     let x = vCast v
     rand <- liftIO $ randomRIO (0, if x == 0 then 1 else x)
@@ -81,6 +88,11 @@ op1 "say" = \v -> do
     op1 "print" v
     liftIO $ putStrLn ""
     return $ VBool True
+op1 "join"= \v -> do
+    v <- readMVal v
+    vals <- mapM readMVal (vCast v)
+    let (pivot:list) = map vCast vals
+    return $ VStr $ concat $ intersperse pivot list
 op1 "die" = \v -> do
     return $ VError (concatMap vCast . vCast $ v) (Val v)
 op1 "exit" = \v -> do
@@ -88,12 +100,26 @@ op1 "exit" = \v -> do
         then liftIO $ exitWith (ExitFailure $ vCast v)
         else liftIO $ exitWith ExitSuccess
 -- handle timely destruction
+op1 "readlink" = \v -> do
+    file <- liftIO $ catch (readSymbolicLink (vCast v)) (\_ -> return "")
+    if file == ""
+      then return VUndef
+      else return $ VStr file
+op1 "sleep" = boolIO sleep
+op1 "mkdir" = boolIO createDirectory
+op1 "rmdir" = boolIO removeDirectory
+op1 "chdir" = boolIO setCurrentDirectory
 op1 "open" = \v -> do
     fh <- liftIO $ openFile (vCast v) ReadMode
     return $ VHandle fh
-op1 "close" = \v -> do
-    liftIO $ hClose (vCast v)
-    return $ VBool True
+op1 "close" = boolIO hClose
+op1 "key" = return . fst . (vCast :: Val -> VPair)
+op1 "value" = return . snd . (vCast :: Val -> VPair)
+op1 "kv" = \v -> do
+    let pair = vCast v
+    return $ VList [fst pair, snd pair]
+op1 "keys" = return . VList . map fst . (vCast :: Val -> [VPair])
+op1 "values" = return . VList . map snd . (vCast :: Val -> [VPair])
 op1 "<>" = \v -> do
     str <- readFrom v
     cxt <- asks envContext
@@ -103,7 +129,7 @@ op1 "<>" = \v -> do
     where
     readFrom VUndef = do
         -- ARGS etc
-        glob <- asks envGlobal
+        glob <- askGlobal
         strs <- liftIO $ sequence $ case find ((== "@*ARGS") . symName) glob of
             Nothing     -> [getStdin glob]
             Just sym    -> case symExp sym of
@@ -122,6 +148,18 @@ op1 ""     = return . (\x -> VError ("unimplemented unaryOp: " ++ "") (Val x))
 
 op1 s      = return . (\x -> VError ("unimplemented unaryOp: " ++ s) (Val x))
 
+boolIO f v = do
+    ok <- liftIO $ (`catch` \_ -> return False) $ do
+        f (vCast v)
+        return True
+    return $ VBool ok
+
+boolIO2 f u v = do
+    ok <- liftIO $ (`catch` \_ -> return False) $ do
+        f (vCast u) (vCast v)
+        return True
+    return $ VBool ok
+
 opEval :: String -> Eval Val
 opEval str = do
     env <- ask
@@ -139,18 +177,21 @@ mapStr f = map (chr . fromEnum . f)
 mapStr2 :: (Word8 -> Word8 -> Word8) -> [Word8] -> [Word8] -> String
 mapStr2 f x y = map (chr . fromEnum . uncurry f) $ x `zip` y
 
-op2 :: Ident -> Val -> Val -> Val
+op2 :: Ident -> Val -> Val -> Eval Val
+op2 "rename" = boolIO2 rename
+op2 "symlink" = boolIO2 createSymbolicLink
+op2 "link" = boolIO2 createLink
 op2 "*"  = op2Numeric (*)
 op2 "/"  = op2Divide
 op2 "%"  = op2Int mod
-op2 "x"  = \x y -> VStr (concat $ (vCast y :: VInt) `genericReplicate` (vCast x :: VStr))
-op2 "xx" = \x y -> VList ((vCast y :: VInt) `genericReplicate` x)
+op2 "x"  = \x y -> return $ VStr (concat $ (vCast y :: VInt) `genericReplicate` (vCast x :: VStr))
+op2 "xx" = \x y -> return $ VList ((vCast y :: VInt) `genericReplicate` x)
 op2 "+&" = op2Int (.&.)
-op2 "+<<"= op2Int shiftL
-op2 "+>>"= op2Int shiftR
+op2 "+<" = op2Int shiftL
+op2 "+>" = op2Int shiftR
 op2 "~&" = op2Str $ mapStr2 (.&.)
-op2 "~<<"= \x y -> VStr $ mapStr (`shiftL` vCast y) (vCast x)
-op2 "~>>"= \x y -> VStr $ mapStr (`shiftR` vCast y) (vCast x)
+op2 "~<" = \x y -> return $ VStr $ mapStr (`shiftL` vCast y) (vCast x)
+op2 "~>" = \x y -> return $ VStr $ mapStr (`shiftR` vCast y) (vCast x)
 op2 "**" = op2Rat ((^^) :: VRat -> VInt -> VRat)
 op2 "+"  = op2Numeric (+)
 op2 "-"  = op2Numeric (-)
@@ -159,7 +200,7 @@ op2 "+|" = op2Int (.|.)
 op2 "+^" = op2Int xor
 op2 "~|" = op2Str $ mapStr2 (.|.)
 op2 "~^" = op2Str $ mapStr2 xor
-op2 "=>" = VPair
+op2 "=>" = \x y -> return $ VPair (x, y)
 op2 "cmp"= op2Ord vCastStr
 op2 "<=>"= op2Ord vCastRat
 op2 ".." = op2Range
@@ -186,41 +227,74 @@ op2 "or" = op2 "||"
 op2 "xor"= op2 "^^"
 op2 "err"= op2 "//"
 op2 "nor"= op2 "!!"
-op2 ";"  = \x y -> y -- XXX wrong! LoL!
-op2 s    = \x y -> VError ("unimplemented binaryOp: " ++ s) (App s [] [Val x, Val y])
+op2 "grep"= op2Grep
+op2 "map"= op2Map
+op2 "split"= \x y -> return $ split (vCast x) (vCast y)
+    where
+    split :: VStr -> VStr -> Val
+    split [] xs = VList $ map (VStr . (:[])) xs
+    split glue xs = VList $ map VStr $ split' xs
+	where
+	split' [] = []
+	split' xs = piece : split' (dropGlue rest)
+	    where (piece, rest) = breakOnGlue glue xs
+	dropGlue = drop (length glue)
+    breakOnGlue _ [] = ([],[])
+    breakOnGlue glue rest@(x:xs)
+	| glue `isPrefixOf` rest = ([], rest)
+	| otherwise = (x:piece, rest') where (piece, rest') = breakOnGlue glue xs
+op2 s    = \x y -> return $ VError ("unimplemented binaryOp: " ++ s) (App s [] [Val x, Val y])
+
+op2Grep list sub@(VSub _) = op2Grep sub list
+op2Grep sub list = do
+    vals <- (`filterM` vCast list) $ \x -> do
+        evl <- asks envEval
+        rv  <- local (\e -> e{ envContext = "Bool" }) $ do
+            evl (Syn "()" [Val sub, Syn "invs" [Val x], Syn "args" []])
+        return $ vCast rv
+    return $ VList vals
+
+op2Map list sub@(VSub _) = op2Map sub list
+op2Map sub list = do
+    vals <- (`mapM` vCast list) $ \x -> do
+        evl <- asks envEval
+        rv  <- local (\e -> e{ envContext = "List" }) $ do
+            evl (Syn "()" [Val sub, Syn "invs" [Val x], Syn "args" []])
+        return $ vCast rv
+    return $ VList $ concat vals
 
 vCastStr :: Val -> VStr
 vCastStr = vCast
 vCastRat :: Val -> VRat
 vCastRat = vCast
 
-op2Str  f x y = VStr  $ f (vCast x) (vCast y)
-op2Num  f x y = VNum  $ f (vCast x) (vCast y)
-op2Rat  f x y = VRat  $ f (vCast x) (vCast y)
-op2Bool f x y = VBool $ f (vCast x) (vCast y)
-op2Int  f x y = VInt  $ f (vCast x) (vCast y)
+op2Str  f x y = return $ VStr  $ f (vCast x) (vCast y)
+op2Num  f x y = return $ VNum  $ f (vCast x) (vCast y)
+op2Rat  f x y = return $ VRat  $ f (vCast x) (vCast y)
+op2Bool f x y = return $ VBool $ f (vCast x) (vCast y)
+op2Int  f x y = return $ VInt  $ f (vCast x) (vCast y)
 
 op1Range (VStr s)    = VList $ map VStr $ strRangeInf s
 op1Range (VRat n)    = VList $ map VRat [n ..]
 op1Range (VNum n)    = VList $ map VNum [n ..]
 op1Range x           = VList $ map VInt [vCast x ..]
 
-op2Range (VStr s) y  = VList $ map VStr $ strRange s (vCast y)
-op2Range (VNum n) y  = VList $ map VNum [n .. vCast y]
-op2Range x (VNum n)  = VList $ map VNum [vCast x .. n]
-op2Range (VRat n) y  = VList $ map VRat [n .. vCast y]
-op2Range x (VRat n)  = VList $ map VRat [vCast x .. n]
-op2Range x y         = VList $ map VInt [vCast x .. vCast y]
+op2Range (VStr s) y  = return $ VList $ map VStr $ strRange s (vCast y)
+op2Range (VNum n) y  = return $ VList $ map VNum [n .. vCast y]
+op2Range x (VNum n)  = return $ VList $ map VNum [vCast x .. n]
+op2Range (VRat n) y  = return $ VList $ map VRat [n .. vCast y]
+op2Range x (VRat n)  = return $ VList $ map VRat [vCast x .. n]
+op2Range x y         = return $ VList $ map VInt [vCast x .. vCast y]
 
 op2Divide x y
     | VInt x' <- x, VInt y' <- y
-    = if y' == 0 then err else VRat $ x' % y'
+    = return $ if y' == 0 then err else VRat $ x' % y'
     | VInt x' <- x, VRat y' <- y
-    = if y' == 0 then err else VRat $ (x' % 1) / y'
+    = return $ if y' == 0 then err else VRat $ (x' % 1) / y'
     | VRat x' <- x, VInt y' <- y
-    = if y' == 0 then err else VRat $ x' / (y' % 1)
+    = return $ if y' == 0 then err else VRat $ x' / (y' % 1)
     | VRat x' <- x, VRat y' <- y
-    = if y' == 0 then err else VRat $ x' / y'
+    = return $ if y' == 0 then err else VRat $ x' / y'
     | otherwise
     = op2Num (/) x y
     where
@@ -232,13 +306,13 @@ op2ChainedList x y
     | VList ys <- y                 = VList (x:ys)
     | otherwise                     = VList [x, y]
 
-op2Logical f x y = if f (vCast x) then x else y
+op2Logical f x y = return $ if f (vCast x) then x else y
 
 op2DefinedOr = undefined
 
-op2Cmp f cmp x y = VBool $ f x `cmp` f y
+op2Cmp f cmp x y = return $ VBool $ f x `cmp` f y
 
-op2Ord f x y = VInt $ case f x `compare` f y of
+op2Ord f x y = return $ VInt $ case f x `compare` f y of
     LT -> -1
     EQ -> 0
     GT -> 1
@@ -253,13 +327,13 @@ op1Numeric f (VRat x)   = VRat $ f x
 op1Numeric f x          = VNum $ f (vCast x)
 
 --- XXX wrong: try num first, then int, then vcast to Rat (I think)
-op2Numeric :: (forall a. (Num a) => a -> a -> a) -> Val -> Val -> Val
+op2Numeric :: (forall a. (Num a) => a -> a -> a) -> Val -> Val -> Eval Val
 op2Numeric f x y
-    | (VInt x', VInt y') <- (x, y)  = VInt $ f x' y'
-    | (VRat x', VInt y') <- (x, y)  = VRat $ f x' (y' % 1)
-    | (VInt x', VRat y') <- (x, y)  = VRat $ f (x' % 1) y'
-    | (VRat x', VRat y') <- (x, y)  = VRat $ f x' y'
-    | otherwise                     = VNum $ f (vCast x) (vCast y)
+    | (VInt x', VInt y') <- (x, y)  = return $ VInt $ f x' y'
+    | (VRat x', VInt y') <- (x, y)  = return $ VRat $ f x' (y' % 1)
+    | (VInt x', VRat y') <- (x, y)  = return $ VRat $ f (x' % 1) y'
+    | (VRat x', VRat y') <- (x, y)  = return $ VRat $ f x' y'
+    | otherwise                     = return $ VNum $ f (vCast x) (vCast y)
 
 primOp :: String -> String -> Params -> String -> Symbol
 primOp sym assoc prms ret = Symbol SOur name (Val sub)
@@ -276,25 +350,30 @@ primOp sym assoc prms ret = Symbol SOur name (Val sub)
                       }
     f :: [Val] -> Eval Val
     f    = case arity of
-        0 -> \(x:_) -> return $ op0 sym (vCast x)
-        1 -> \[x]   -> op1 sym (vCast x)
-        2 -> \[x,y] -> return $ op2 sym (vCast x) (vCast y)
+        0 -> \(x:_) -> op0 sym (vCast x)
+        1 -> \x     -> case x of
+            [x]   -> op1 symName x
+            [x,y] -> op2 sym x y
+            x     -> op0 sym x
+        2 -> \[x,y] -> op2 sym (vCast x) (vCast y)
         _ -> error (show arity)
-    (arity, fixity) = case assoc of
-        "pre"       -> (1, "prefix")
-        "post"      -> (1, "postfix")
-        "circum"    -> (1, "circumfix")
-        "left"      -> (2, "infix")
-        "right"     -> (2, "infix")
-        "non"       -> (2, "infix")
-        "chain"     -> (2, "infix")
-        "list"      -> (0, "infix")
-        other       -> (0, other)
+    symName = if modify then assoc ++ ":" ++ sym else sym
+    (arity, fixity, modify) = case assoc of
+        "pre"       -> (1, "prefix", False)
+        "post"      -> (1, "postfix", True)
+        "circum"    -> (1, "circumfix", True)
+        "left"      -> (2, "infix", False)
+        "right"     -> (2, "infix", False)
+        "non"       -> (2, "infix", False)
+        "chain"     -> (2, "infix", False)
+        "list"      -> (0, "infix", False)
+        other       -> (0, other, True)
         
 primDecl str = primOp sym assoc (reverse $ foldr foldParam [] prms) ret
     where
     (ret:assoc:sym:prms') = words str
-    takeWord = takeWhile isAlphaNum . dropWhile (not . isAlphaNum)
+    takeWord = takeWhile isWord . dropWhile (not . isWord)
+    isWord = not . (`elem` "|(),")
     prms = map takeWord prms'
 
 doFoldParam cxt [] []       = [buildParam cxt "" "$?1" (Val VUndef)]
@@ -303,7 +382,11 @@ doFoldParam cxt (s:name) ps = (buildParam cxt [s] name (Val VUndef) : ps)
 
 foldParam :: String -> Params -> Params
 foldParam "List"    = doFoldParam "List" "*@?0"
-foldParam "?Num=1"  = \ps -> (buildParam "Num" "?" "$?1" (Val $ VNum 1):ps)
+foldParam ('r':'w':'!':str) = \ps -> ((buildParam str "" "$?1" (Val VUndef)) { isLValue = True }:ps)
+foldParam ""        = id
+foldParam ('?':str) = \ps -> (buildParam "Num" "?" "$?1" (Val $ VNum (read def)):ps)
+    where
+    (cxt, ('=':def)) = break (== '=') str
 foldParam x         = doFoldParam x ""
 
 -- XXX -- Junctive Types -- XXX --
@@ -321,13 +404,24 @@ initSyms = map primDecl . filter (not . null) . lines $ "\
 \\n   Bool      pre     ?^      (Bool)\
 \\n   Ref       pre     \\      (Any)\
 \\n   List      pre     ...     (Str|Num)\
-\\n   Any       post    ++      (LValue)\
-\\n   Num       post    --      (LValue)\
+\\n   Any       post    ++      (rw!Num)\
+\\n   Num       post    --      (rw!Num)\
+\\n   Any       pre     ++      (rw!Num)\
+\\n   Num       pre     --      (rw!Num)\
 \\n   Bool      pre     not     (Bool)\
+\\n   List      pre     map     (Code, List)\
+\\n   List      pre     grep    (Code, List)\
+\\n   List      pre     map     (List: Code)\
+\\n   List      pre     grep    (List: Code)\
+\\n   Str       pre     join    (List)\
+\\n   List      left    zip     (List)\
+\\n   List      pre     keys    (Hash)\
+\\n   List      pre     values  (Hash)\
+\\n   List      pre     kv      (Hash)\
 \\n   Str       pre     perl    (List)\
 \\n   Any       pre     eval    (Str)\
-\\n   Any       pre     last    (Num)\
-\\n   Any       pre     exit    (Num)\
+\\n   Any       pre     last    (?Num=1)\
+\\n   Any       pre     exit    (?Num=0)\
 \\n   Num       pre     rand    (?Num=1)\
 \\n   Num       pre     time    ()\
 \\n   Action    pre     print   (List)\
@@ -341,6 +435,22 @@ initSyms = map primDecl . filter (not . null) . lines $ "\
 \\n   Junction  pre     all     (List)\
 \\n   Junction  pre     one     (List)\
 \\n   Junction  pre     none    (List)\
+\\n   Bool      pre     sleep   (Int)\
+\\n   Bool      pre     rmdir   (Str)\
+\\n   Bool      pre     mkdir   (Str)\
+\\n   Bool      pre     chdir   (Str)\
+\\n   Scalar    pre     key     (Pair)\
+\\n   Scalar    pre     value   (Pair)\
+\\n   List      pre     kv      (Pair)\
+\\n   List      pre     keys    (Hash)\
+\\n   List      pre     values  (Hash)\
+\\n   Bool      pre     rename  (Str, Str)\
+\\n   Bool      pre     symlink (Str, Str)\
+\\n   Bool      pre     link    (Str, Str)\
+\\n   Str       pre     readlink (Str)\
+\\n   List      pre     split   (Str, Str)\
+\\n   Str       pre     =       (IO)\
+\\n   List      pre     =       (IO)\
 \\n   Junction  list    |       (List)\
 \\n   Junction  list    &       (List)\
 \\n   Junction  list    ^       (List)\
@@ -351,11 +461,11 @@ initSyms = map primDecl . filter (not . null) . lines $ "\
 \\n   Str       left    x       (Str, Int)\
 \\n   List      left    xx      (Any, Int)\
 \\n   Int       left    +&      (Int, Int)\
-\\n   Int       left    +<<     (Int, Int)\
-\\n   Int       left    +>>     (Int, Int)\
+\\n   Int       left    +<      (Int, Int)\
+\\n   Int       left    +>      (Int, Int)\
 \\n   Str       left    ~&      (Str, Str)\
-\\n   Str       left    ~<<     (Str, Str)\
-\\n   Str       left    ~>>     (Str, Str)\
+\\n   Str       left    ~<      (Str, Str)\
+\\n   Str       left    ~>      (Str, Str)\
 \\n   Num       right   **      (Num, Num)\
 \\n   Num       left    +       (Num, Num)\
 \\n   Num       left    -       (Num, Num)\
